#!/usr/bin/perl -w

#*******************************************************************************
# bf_megane.pl
#
# Brainf*ck の眼鏡版。
# 注意：ソースコードは UTF8 で記載して下さい。
#
#【引数】
#　　ARGV[0]　実行するソースファイルパス。
# 　　　　　　　　　　　　　　　　　　　　　　　　　　　2013/02/19 made   by ktr
#*******************************************************************************

#-------------------------------------------------------------------------------
# モジュール宣言
#-------------------------------------------------------------------------------
use strict;
use warnings;
use utf8;

#-------------------------------------------------------------------------------
# 定数宣言
#-------------------------------------------------------------------------------

# トークンを Brainf*ck のコマンドに置き換えるためのテーブル
my $TOKEN_TABLE = {
	'メガネ'       => '>' ,
	'め・が・ね！' => '<' ,
	'眼鏡'         => '+' ,
	'めがね'       => '-' ,
	'っ娘'         => '.' ,
	'メ・ガ・ネ！' => ',' ,
	'さん'         => '[' ,
	'っ子'         => ']' ,
};


# 処理対象のバイト配列
my @bytes;
# バイト配列の位置を示すポインタ
my $ptr;


# メイン関数呼び出し
&main();

#-------------------------------------------------------------------------------
# メイン処理。
#-------------------------------------------------------------------------------
sub main(){

	# 実行するソースファイルの取得
	my $source_file = $ARGV[0];

	# ソースファイル指定不正時はエラー
	if ( ! -f $source_file ) {
		print "Can't Read File $source_file.\n";
	}

	# ソースファイルの読み込み
	open( IN , "<:utf8" , $source_file  );
	local $/ = undef;
	my $code = <IN>;
	close( IN );

	# ソースコードの正規化
	$code =~ s/0x0d//g;
	$code =~ s/0x0a//g;

	# トークン置換用にテーブルのキーを文字数の多い順に取得
	# ※長い文字から置換することでトークン同士の部分一致による誤変換を回避
	my @token_keys = sort { length $b <=> length $a } keys(%{$TOKEN_TABLE});
	# トークンを Brainf*ck の標準的なコマンドに置換する
	foreach my $token_key( @token_keys ) {
		$code =~ s/$token_key/$TOKEN_TABLE->{$token_key}/g;
	}

	# 余計な文字の削除
	$code =~ s/[^><\+\-\.\,\[\]]//g;

	# ポインタの初期化
	$ptr = 0;

	# コードを解析して実行する
	&interpret( $code );
}

#-------------------------------------------------------------------------------
# コードを解析して実行する。
# @param 解析するコード。
#-------------------------------------------------------------------------------
sub interpret($) {

	# 引数の取得
	my $code = shift;

	# 一文字ずつに分割する
	my @chars = split( // , $code );

	# 一文字ずつ解析
	my $index = 0;
	while ( $index < @chars ) {
		
		# 現在位置の文字を取得
		my $c = $chars[$index];
		
		if ( $c eq '>' ) {
			# ポインタ加算
			$ptr++
		} elsif ( $c eq '<' ) {
			# ポインタ減算
			$ptr--;
		} elsif ( $c eq '+' ) {
			# ポインタの示す値を加算
			$bytes[$ptr]++;
		} elsif ( $c eq '-' ) {
			# ポインタの示す値を減算
			$bytes[$ptr]--;
		} elsif ( $c eq '.' ) {
			# ポインタの示す値を出力
			print chr( $bytes[$ptr] );
		} elsif ( $c eq ',' ) {
			# ポインタの示す位置へ入力
			$bytes[$ptr] = getc();
		} elsif ( $c eq '[' ) {
			# 終了位置の取得
			my $end_index = &search_end_bracket_index( $index + 1 , \@chars );
			# 現在のポインタ位置の保存
			my $now_ptr = $ptr;
			while ( $bytes[$now_ptr] > 0 ) {
				# ブラケット内のコードに対して解析処理を再帰呼び出し
				my $tmp_code;
				my $tmp_index = $index + 1;
				while ( $tmp_index < $end_index ) {
					$tmp_code .= $chars[$tmp_index];
					$tmp_index++;
				}
				&interpret( $tmp_code );
			}
			# ポインタを元の位置に戻す
			$ptr = $now_ptr;
			# インデックスをブラケットの最後に設定する
			$index = $end_index;
		}
		# インデックスを加算する
		$index++;
	}
}
#-------------------------------------------------------------------------------
# 対応するブラケットのインデックスを検索する。
# @param 検索開始インデックス。
# @param 検索対象配列のリファレンス。
#-------------------------------------------------------------------------------
sub search_end_bracket_index($$) {

	# 引数の取得
	my ( $start_index , $array_ref )  = @_;

	# デリファレンスして配列を取得
	my @chars = @{$array_ref};

	# 入れ子に対応する為の深さ
	my $depth = 0;

	# 検索開始インデックスの設定
	my $index = $start_index;
	
	# 検索対象配列内をループ
	while ( $index < @chars ) {

		# 現在位置の文字取得
		my $c = $chars[$index];

		# 深さ 0 で ] が現れた場合はそこがブラケット終了位置
		if ( $c eq ']' && $depth == 0 ) {
			last;
		}

		# [ が現れた場合は深さを加算
		if ( $c eq '[' ) {
			$depth++;

		# ] が現れた場合は深さを減算
		} elsif ( $c eq ']' ) {
			$depth--;
		}
		
		# インデックス加算
		$index++;
	}
	
	# 検索結果のインデックスを返す
	return $index;
}
